/**********************************************************************
 * Programaci�n Concurrente y Distribuida
 * Ingenier�a Inform�tica - Universidad de Cantabria
 * 
 * Proyecto: ActiveDigitalIO Driver
 * 
 * Fichero: Clase DigitalIOCard
 * 
 * @Autor: J.M. Drake y L. Barros
 * @Fecha: 1-11-07
 * 
 *********************************************************************/
import java.util.ArrayList;
import java.util.Iterator;

import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.layout.RowData;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.layout.FillLayout;

/**
 * Ofrece los m�todos est�ticos primitivos de acceso a los registros que controlan las lineas 
 * digitales de entrada y salida.
 */

public class DigitalIOCardCruce extends Thread implements DigitalIOCard{
	
	//	Lista con los observadores DigitalIOListener que se han subscrito
	//  a la tarjeta.
	private ArrayList<DigitalIOListener> listeners=new ArrayList<DigitalIOListener>();

	// Constructor
	public DigitalIOCardCruce(){
		super();
		this.setDaemon(true);
		this.start();

		try {
			Thread.sleep(1000);
		} catch (InterruptedException e) {}
	}	
	
	
	// Public API

	public long readInputBuffer(){
		//static long reg;
		 display.syncExec(new Runnable() {
			public void run() {
				reg=0;
				for (int i=0;i<32;i++){
					if (cb[i].getSelection()) reg =reg+(1<<i);
				}
			}
		});
        return reg;
	}
	
	public void writeOutputBuffer(long data){
		final long theData=data;
		display.syncExec(new Runnable() {
				public void run() {
					for (int i=0;i<32;i++){
						if ((theData&(1L<<i))!=0) {
							rb[i].setSelection(true);
							if (i==0) setCircle(SemaphoreType.SEM_1, ColorType.RED, true);
							if (i==1) setCircle(SemaphoreType.SEM_1, ColorType.ORANGE, true);
							if (i==2) setCircle(SemaphoreType.SEM_1, ColorType.GREEN, true);
							if (i==3) setCircle(SemaphoreType.SEM_2, ColorType.RED, true);
							if (i==4) setCircle(SemaphoreType.SEM_2, ColorType.ORANGE, true);
							if (i==5) setCircle(SemaphoreType.SEM_2, ColorType.GREEN, true);
							if (i==6) setCircle(SemaphoreType.SEM_3, ColorType.RED, true);
							if (i==7) setCircle(SemaphoreType.SEM_3, ColorType.ORANGE, true);
							if (i==8) setCircle(SemaphoreType.SEM_3, ColorType.GREEN, true);
							if (i==9) setFigure(PedestrianSemaphoreType.PED_1, ActionType.STANDING, true);
							if (i==10) setFigure(PedestrianSemaphoreType.PED_1, ActionType.WALKING, true);
							if (i==11) setFigure(PedestrianSemaphoreType.PED_2, ActionType.STANDING, true);
							if (i==12) setFigure(PedestrianSemaphoreType.PED_2, ActionType.WALKING, true);
						    if (i<14) cb[i].setSelection(false);
						}
						else{
							rb[i].setSelection(false);
							if (i==0) setCircle(SemaphoreType.SEM_1, ColorType.RED, false);
							if (i==1) setCircle(SemaphoreType.SEM_1, ColorType.ORANGE, false);
							if (i==2) setCircle(SemaphoreType.SEM_1, ColorType.GREEN, false);
							if (i==3) setCircle(SemaphoreType.SEM_2, ColorType.RED, false);
							if (i==4) setCircle(SemaphoreType.SEM_2, ColorType.ORANGE, false);
							if (i==5) setCircle(SemaphoreType.SEM_2, ColorType.GREEN, false);
							if (i==6) setCircle(SemaphoreType.SEM_3, ColorType.RED, false);
							if (i==7) setCircle(SemaphoreType.SEM_3, ColorType.ORANGE, false);
							if (i==8) setCircle(SemaphoreType.SEM_3, ColorType.GREEN, false);
							if (i==9) setFigure(PedestrianSemaphoreType.PED_1, ActionType.STANDING, false);
							if (i==10) setFigure(PedestrianSemaphoreType.PED_1, ActionType.WALKING, false);
							if (i==11) setFigure(PedestrianSemaphoreType.PED_2, ActionType.STANDING, false);
							if (i==12) setFigure(PedestrianSemaphoreType.PED_2, ActionType.WALKING, false);
							if (i<14) cb[i].setSelection(true);
						}
					}
				}
		 });
		this.notifyInterrupt();
	}
	
	// A�ade el objeto listener que se pasa como par�metro como observa-
	// dor de los cambios que se producen en las entradas digitales de 
	// la tarjeta.
	public void addInterruptListener(DigitalIOListener listener){
		listeners.add(listener);
	}
	
	// M�todo privado que ejecuta el m�todo newChange de todos los 
	// observadores subscritos a los cambios de entrada.
	private void notifyInterrupt(){
		Iterator<DigitalIOListener> iter;
		iter=listeners.iterator();
		while (iter.hasNext()){
			iter.next().newChange();
		}
	}
	
	
	
	/**
	 * Recursos de la GUI
	 */

	// Constants for drawing the street
	private final int[] ORIGIN = {10,40};
	private final int WIDTH = 665;
	private final int ROAD_WIDTH = 200;
	private final int LINE_WIDTH = ROAD_WIDTH / 50;
	private final int LINE_SEGMENT = 10;
	private final int LINE_CROSS = 6*LINE_SEGMENT;
	private final int RADIO = 20;
	private final int SEMAPHORE_HEIGHT = 75 + RADIO/4; 
	private final int SEMAPHORE_WIDTH = SEMAPHORE_HEIGHT / 3 + RADIO/4;
	
	private enum SemaphoreType {SEM_1, SEM_3, SEM_2}
	private enum ColorType {RED, ORANGE, GREEN}
	private enum PedestrianSemaphoreType {PED_1, PED_2}
	private enum ActionType {STANDING, WALKING}
	
	private SemaphoreInfo semaphore1 = new SemaphoreInfo(ORIGIN[0] + WIDTH/2 + ROAD_WIDTH/4 + LINE_WIDTH, ORIGIN[1] + ROAD_WIDTH + LINE_CROSS + 2*LINE_WIDTH, true);
	private SemaphoreInfo semaphore2 = new SemaphoreInfo(ORIGIN[0] + (2*WIDTH)/3 + LINE_CROSS, ORIGIN[1] - SEMAPHORE_WIDTH - LINE_WIDTH, false);
	private SemaphoreInfo semaphore3 = new SemaphoreInfo(ORIGIN[0] + WIDTH/2 - 2*ROAD_WIDTH/3, ORIGIN[1] + ROAD_WIDTH + LINE_WIDTH, false);
	
	private PedestrianSemaphoreInfo pedestrian1 = new PedestrianSemaphoreInfo(ORIGIN[0] + WIDTH/2 - ROAD_WIDTH/4 - SEMAPHORE_WIDTH - LINE_WIDTH, ORIGIN[1] + ROAD_WIDTH + SEMAPHORE_WIDTH + 2*LINE_WIDTH);
	private PedestrianSemaphoreInfo pedestrian2 = new PedestrianSemaphoreInfo(ORIGIN[0] + (3*WIDTH)/4, ORIGIN[1] + ROAD_WIDTH + LINE_WIDTH);
	
	// GUI elements
	Display display = null;  
	private GC gc = null;
	private Shell sShell = null; 
	private Composite c_buttons = null;
	private Canvas cv_street = null;
	
	private static Button[] rb = new Button[32];
	private static Button[] cb = new Button[32];
	
	static long reg;
	
	
	public void run(){
    	display = Display.getDefault();
		createSShell();
		sShell.open();
		drawStreet();
	    
	    writeOutputBuffer(0);

		while (!sShell.isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}
		display.dispose();
    }
	
	/**
	 * This method initializes sShell	
	 *
	 */
	private void createSShell() {
		
		RowLayout rowLayout = new RowLayout();
			rowLayout.type = org.eclipse.swt.SWT.VERTICAL;
		sShell = new Shell();
			sShell.setText("IOCard visor");
			sShell.setLayout(rowLayout);
			sShell.setSize(new Point(700, 530));
		createCv_road();
		createC_buttons();			
	}
	
	/**
	 * This method initializes c_buttons	
	 *
	 */
	private void createC_buttons() {
		
		GridLayout gridLayout = new GridLayout();
			gridLayout.numColumns = 32;
			gridLayout.marginHeight = 30;
		RowData rowData = new RowData();
			rowData.height = 100;
			rowData.width = 685;
		c_buttons = new Composite(sShell, SWT.NONE);
			c_buttons.setLayoutData(rowData);
			c_buttons.setLayout(gridLayout);
			c_buttons.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_WIDGET_BACKGROUND));
			
		for (int i = 31; i >= 0; i--){
			GridData gridData = new GridData();
				gridData.grabExcessHorizontalSpace = true;
			rb[i] = new Button(c_buttons,SWT.RADIO|SWT.NO_RADIO_GROUP);
				rb[i].setLayoutData(gridData);
				rb[i].setEnabled(false);
		}
		for (int i = 31; i >= 0; i--){
			cb[i] = new Button(c_buttons,SWT.CHECK);
			if (i<30) cb[i].setEnabled(false);
			cb[i].addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {  //NUEVO
				public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
					notifyInterrupt();
				}
			});
		}
		
		cb[30].addSelectionListener(new SelectionAdapter(){

			public void widgetSelected(SelectionEvent e) {
				if (cb[30].getSelection()) {
					gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));						
				}else {
					gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));						
				}
				gc.fillOval(ORIGIN[0] + (2*WIDTH)/3 + LINE_CROSS + SEMAPHORE_WIDTH + RADIO/8, ORIGIN[1] + ROAD_WIDTH + SEMAPHORE_WIDTH + RADIO/8, RADIO/2, RADIO/2);
			}
		});
		
		cb[31].addSelectionListener(new SelectionAdapter(){

			public void widgetSelected(SelectionEvent e) {
				if (cb[31].getSelection()) {
					gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_RED));					
				}else {
					gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));						
				}
				gc.fillRectangle(ORIGIN[0] + WIDTH/2 - ROAD_WIDTH/8 + LINE_WIDTH, ORIGIN[1] + ROAD_WIDTH + 3*LINE_CROSS/2 + LINE_WIDTH, ROAD_WIDTH/4 - 2*LINE_WIDTH + 1, LINE_WIDTH + 1);
			}
		});
	}

	/**
	 * This method initializes cv_road	
	 *
	 */
	private void createCv_road() {
		RowData rowData1 = new RowData();
			rowData1.height = 385;
			rowData1.width = 685;
		cv_street = new Canvas(sShell, SWT.NONE);
			cv_street.setLayout(new FillLayout());
			cv_street.setLayoutData(rowData1);
			cv_street.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_DARK_GREEN));
	}
	
	
	private void drawStreet(){
		
		gc = new GC(cv_street);
		
		// I draw the roads
		gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_DARK_GRAY));
			gc.fillRectangle(ORIGIN[0], ORIGIN[1], WIDTH + 1, ROAD_WIDTH + 1);
			gc.fillRectangle(ORIGIN[0] + WIDTH/2 - ROAD_WIDTH/4, ORIGIN[1] + ROAD_WIDTH, ROAD_WIDTH/2 + 1, ROAD_WIDTH + 1);
		
		// I draw the white lines
		gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
			
			// central continous line
			gc.fillRectangle(ORIGIN[0], ORIGIN[1] + ROAD_WIDTH/2 - LINE_WIDTH/2, WIDTH + 1, LINE_WIDTH + 1);
			
			// upper discontinous line
			for (int i = 0; i < WIDTH / (2*LINE_SEGMENT); i++) {
				gc.fillRectangle(ORIGIN[0] + LINE_SEGMENT/2 + i*2*LINE_SEGMENT, ORIGIN[1] + ROAD_WIDTH/4, LINE_SEGMENT + 1, LINE_WIDTH + 1);
			}
			
			// lower discontinous line
			for (int i = 0; i < WIDTH / (2*LINE_SEGMENT); i++) {
				gc.fillRectangle(ORIGIN[0]+ LINE_SEGMENT/2 + i*2*LINE_SEGMENT, ORIGIN[1] + 3*ROAD_WIDTH/4 - LINE_WIDTH, LINE_SEGMENT + 1, LINE_WIDTH + 1);
			}	
			
			// main road's pedestrian cross
			for (int i = 0; i < ROAD_WIDTH / (2*LINE_WIDTH); i++) {
				gc.fillRectangle(ORIGIN[0] + (2*WIDTH)/3, ORIGIN[1] + LINE_WIDTH/2 + i*2*LINE_WIDTH, LINE_CROSS + 1, LINE_WIDTH + 1);
			}
			
			// secondary road's pedestrian cross
			for (int i = 0; i < (ROAD_WIDTH/2) / (2*LINE_WIDTH); i++) {
				gc.fillRectangle(ORIGIN[0] + WIDTH/2 - ROAD_WIDTH/4 + LINE_WIDTH + i*2*LINE_WIDTH, ORIGIN[1] + ROAD_WIDTH + LINE_SEGMENT, LINE_WIDTH + 1, LINE_CROSS + 1);
			}
			
			// arrows
			Image arrow = new Image(Display.getDefault(),"./arrow.gif");
			gc.drawImage(arrow, ORIGIN[0] + ROAD_WIDTH/2, ORIGIN[1] + 5*ROAD_WIDTH/8 - 10);
			gc.drawImage(arrow, ORIGIN[0] + ROAD_WIDTH/2, ORIGIN[1] + 7*ROAD_WIDTH/8 - 10);
			arrow.dispose();
			Image arrow2 = new Image(Display.getDefault(),"./arrow2.gif");
			gc.drawImage(arrow2, ORIGIN[0] + WIDTH - ROAD_WIDTH/2, ORIGIN[1] + ROAD_WIDTH/8 - 10);
			gc.drawImage(arrow2, ORIGIN[0] + WIDTH - ROAD_WIDTH/2, ORIGIN[1] + 3*ROAD_WIDTH/8 - 10);
			arrow2.dispose();
			
		// I draw the discontinuous central line
		gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_DARK_GRAY));
			for (int i = 0; i < (ROAD_WIDTH/2)/LINE_SEGMENT; i++) {
				gc.fillRectangle(ORIGIN[0] + WIDTH/2 - ROAD_WIDTH/2 + i*LINE_SEGMENT, ORIGIN[1] + ROAD_WIDTH/2 - LINE_WIDTH/2, LINE_SEGMENT/2 + 1, LINE_WIDTH + 1);
			}
		
		// I draw the 3 semaphores
		drawSemaphore(SemaphoreType.SEM_1);
		drawSemaphore(SemaphoreType.SEM_2);			
		drawSemaphore(SemaphoreType.SEM_3);
		
		// I draw the 2 pedestrian semaphores
		drawPedestrianSemaphore(PedestrianSemaphoreType.PED_1);
		drawPedestrianSemaphore(PedestrianSemaphoreType.PED_2);
		
		// I draw the small point
		gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
			gc.fillOval(ORIGIN[0] + (2*WIDTH)/3 + LINE_CROSS + SEMAPHORE_WIDTH, ORIGIN[1] + ROAD_WIDTH + SEMAPHORE_WIDTH, 3*RADIO/4, 3*RADIO/4);
		
		// I draw the horizontal line
		gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
			gc.fillRectangle(ORIGIN[0] + WIDTH/2 - ROAD_WIDTH/8, ORIGIN[1] + ROAD_WIDTH + 3*LINE_CROSS/2, ROAD_WIDTH/4, 3*LINE_WIDTH);
		
	}
	
	
	private SemaphoreInfo getSemaphoreInfo(SemaphoreType semID){
		
		switch (semID) {
			case SEM_1:
				return semaphore1;
			case SEM_2:
				return semaphore2;
			case SEM_3:
				return semaphore3;
			default:
				return null;
		}
	}
	
	private void drawSemaphore(SemaphoreType semID){
		
		SemaphoreInfo semaphore = getSemaphoreInfo(semID);
		
		gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
		
		if (semaphore.vertical) {
			gc.fillRectangle(semaphore.x, semaphore.y, SEMAPHORE_WIDTH + 1, SEMAPHORE_HEIGHT + 1);
				setCircle(semID, ColorType.RED, true);
				setCircle(semID, ColorType.ORANGE, true);						
				setCircle(semID, ColorType.GREEN, true);
		} else {
			gc.fillRectangle(semaphore.x, semaphore.y, SEMAPHORE_HEIGHT + 1, SEMAPHORE_WIDTH + 1);
				setCircle(semID, ColorType.RED, true);
				setCircle(semID, ColorType.ORANGE, true);
				setCircle(semID, ColorType.GREEN, true);				
		}					
	}
	
	private void setCircle(final SemaphoreType semID, final ColorType colorID, final boolean state){
		display.syncExec(new Runnable(){
			public void run(){
				SemaphoreInfo semaphore = getSemaphoreInfo(semID);
		
				switch (colorID) {
				
					case RED:
						if (state) {
							gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_RED));
						} else {
							gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
						}
						if (semaphore.vertical) {
							gc.fillOval(semaphore.x + RADIO/4, semaphore.y + RADIO/4, RADIO, RADIO);
						} else {
							gc.fillOval(semaphore.x + RADIO/4, semaphore.y + RADIO/4, RADIO, RADIO);
						}
						break;
									
					case ORANGE:
						if (state) {
							gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_YELLOW));
						} else {
							gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
						}				
						if (semaphore.vertical) {
							gc.fillOval(semaphore.x + RADIO/4, semaphore.y + 3*RADIO/2, RADIO, RADIO);
						} else {
							gc.fillOval(semaphore.x + 3*RADIO/2, semaphore.y + RADIO/4, RADIO, RADIO);
						}
						break;
					
					case GREEN:
						if (state) {
							gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_GREEN));
						} else {
							gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
						}				
						if (semaphore.vertical) {
							gc.fillOval(semaphore.x + RADIO/4, semaphore.y  + 11*RADIO/4, RADIO, RADIO);
						} else {
							gc.fillOval(semaphore.x + 11*RADIO/4, semaphore.y + RADIO/4, RADIO, RADIO);
						}
						break;
					default:
						break;
				}
			}
		}
		);
	}
	
	
	private PedestrianSemaphoreInfo getPedestrianSemaphoreInfo(PedestrianSemaphoreType pedID){
		
		switch (pedID) {
			case PED_1:
				return pedestrian1;
			case PED_2:
				return pedestrian2;
			default:
				return null;
		}
	}
	
	private void drawPedestrianSemaphore(PedestrianSemaphoreType pedID){
		
		PedestrianSemaphoreInfo pedestrian = getPedestrianSemaphoreInfo(pedID);
		
		gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
			gc.fillRectangle(pedestrian.x, pedestrian.y, SEMAPHORE_WIDTH, 2*SEMAPHORE_WIDTH);
		gc.setForeground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
			gc.drawLine(pedestrian.x, pedestrian.y + SEMAPHORE_WIDTH, pedestrian.x + SEMAPHORE_WIDTH, pedestrian.y + SEMAPHORE_WIDTH);
		
		setFigure(pedID, ActionType.STANDING, true);
		setFigure(pedID, ActionType.WALKING, true);
	}
	
	private void setFigure(final PedestrianSemaphoreType pedID,final ActionType actionID,final boolean state){
		display.syncExec(new Runnable(){
			public void run(){
				PedestrianSemaphoreInfo pedestrian = getPedestrianSemaphoreInfo(pedID);
		
				switch (actionID) {
			
				case STANDING:
					if (state) {
						Image pedestrianStanding = new Image(Display.getDefault(),"./pedestrianRed.gif");
						gc.drawImage(pedestrianStanding, pedestrian.x + RADIO/4, pedestrian.y + RADIO/4);
						pedestrianStanding.dispose();
					} else {
						gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
						//gc.fillRectangle(pedestrian.x, pedestrian.y, SEMAPHORE_WIDTH, 2*SEMAPHORE_WIDTH);
						gc.fillRectangle(pedestrian.x, pedestrian.y, SEMAPHORE_WIDTH, SEMAPHORE_WIDTH);
					}
					break;
				case WALKING:
					if (state) {
						Image pedestrianWalking = new Image(Display.getDefault(),"./pedestrianGreen.gif");
						gc.drawImage(pedestrianWalking, pedestrian.x + RADIO/4, pedestrian.y + RADIO/4 + SEMAPHORE_WIDTH);
						pedestrianWalking.dispose();
					} else {
						gc.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
						//gc.fillRectangle(pedestrian.x, pedestrian.y, SEMAPHORE_WIDTH, SEMAPHORE_WIDTH);
						gc.fillRectangle(pedestrian.x, pedestrian.y+SEMAPHORE_WIDTH, SEMAPHORE_WIDTH, SEMAPHORE_WIDTH);
					}	
					break;
				default:
					break;
				};
			};
		}
		);
	}

	// Inner classes
	
	private class SemaphoreInfo{
		protected int x;
		protected int y;
		protected boolean vertical;
		
		SemaphoreInfo(int x, int y, boolean vertical){
			this.x = x;
			this.y = y;
			this.vertical = vertical;
		}
	}
	
	private class PedestrianSemaphoreInfo{
		protected int x;
		protected int y;
		
		private PedestrianSemaphoreInfo(int x, int y){
			this.x = x;
			this.y = y;
		};
	}
}
