import CruceSemaforosPkt.DigitalIOCardRmtPrx;
import CruceSemaforosPkt.DigitalIOCardRmtPrxHelper;
import CruceSemaforosPkt.DigitalIODriverRmtPrx;
import CruceSemaforosPkt.DigitalIODriverRmtPrxHelper;
import CruceSemaforosPkt.DigitalIOListenerRmtPrx;
import CruceSemaforosPkt.DigitalIOListenerRmtPrxHelper;
import CruceSemaforosPkt.IndexOutOfRange;
import CruceSemaforosPkt.MeHanInterrumpido;
import CruceSemaforosPkt._DigitalIODriverRmtOperations;
import CruceSemaforosPkt._DigitalIODriverRmtTie;
import CruceSemaforosPkt._DigitalIOListenerRmtOperations;
import CruceSemaforosPkt._DigitalIOListenerRmtTie;

/**********************************************************************
 * Programaci�n Concurrente y Distribuida
 * Ingenier�a Inform�tica - Universidad de Cantabria
 * 
 * Proyecto: Clase DigitalIO Driver
 * 
 * Fichero: Clase ActiveIODigitalDriver
 * 
 * @Autor: J.M. Drake y L.Barros
 * @Fecha: 22-10-08
 * 
 *********************************************************************/

/**
 * Driver que gestiona el acceso concurrente a una tarjeta de entrada/salida
 * digital. Implementa la interfaz DigitalIOListener
 */
public class DigitalIODriver implements _DigitalIOListenerRmtOperations, 
										_DigitalIODriverRmtOperations 
{

	// Estado actual de las lineas digitales de salida. Se requiere como
	// variable de estado del driver porque la tajeta no permite leer el
	// valor que est� establecida en la tarjeta.
	public long outputCurrentState;

	// Referencia la tarjeta de IO Digital que gestiona.
	private DigitalIOCardRmtPrx theCard;

	private final int NUM_LINE = 32;

	OutputGenerator[] generators;
	
	public static void main (String args[]){
		Ice.Properties propiedades = new Ice.Util().createProperties(args);

        propiedades.setProperty("Ice.ThreadPool.Server.Size", "5" );
        propiedades.setProperty("Ice.ThreadPool.Client.Size", "5" );

        Ice.InitializationData init_data = new Ice.InitializationData();
        init_data.properties = propiedades;

        Ice.Communicator ic = Ice.Util.initialize(init_data);
        new DigitalIODriver(ic);
	}

	/**
	 * Constructor de la clase
	 */
	DigitalIODriver(Ice.Communicator ic) 
	{
		this.outputCurrentState = 0;
		
		boolean exito = false;
		
		/* Buscamos la tarjeta */
		/***********************/
		while(exito == false)
		{
			try
			{
				String icestr = DatosComunes.nombreCardRmt + 
							":default -p " + DatosComunes.puerto_tcp_tarjeta;
			
				Ice.ObjectPrx proxy_general = ic.stringToProxy(icestr);
				theCard = DigitalIOCardRmtPrxHelper.checkedCast(proxy_general);
				exito = true;
			}
			catch (Ice.SocketException e)
			{
				System.out.println("DigitalIODriver:  problemas con sockets");
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
			}
			catch (Exception e)
			{
				System.out.println("Nueva exception");
				e.printStackTrace();
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
			}
		
		}
		
		/* Creamos un Ice adapter y agnadimos los servants de DigitalIODriver y */
		/* digitalIOListener.                                                   */
		/************************************************************************/
		
		try
		{
			String icestr =  "default -p " + DatosComunes.puerto_tcp_driver;
			Ice.ObjectAdapter adapter = ic.createObjectAdapterWithEndpoints("AdapterDriver",
											icestr);
			
			// Ties
			_DigitalIODriverRmtTie servant_driver = new _DigitalIODriverRmtTie(this);
			_DigitalIOListenerRmtTie servant_listener = new _DigitalIOListenerRmtTie(this);
			
			Ice.ObjectPrx proxy_general_driver = adapter.add(servant_driver, ic.stringToIdentity(DatosComunes.nombreDriverRmt) );
			Ice.ObjectPrx proxy_general_listener =  adapter.add(servant_listener, ic.stringToIdentity(DatosComunes.nombreListenerRmt) );

			adapter.activate();

			// Obtenemos los proxies y los pasamos a la tarjeta
			DigitalIODriverRmtPrx stub_driver = DigitalIODriverRmtPrxHelper.checkedCast(proxy_general_driver);
			DigitalIOListenerRmtPrx stub_listener = DigitalIOListenerRmtPrxHelper.checkedCast(proxy_general_listener);
			
			theCard.addDigitalIODriver(stub_driver);
			theCard.addInterruptListener(stub_listener);
		}
		catch (Ice.SocketException e)
		{
			System.out.println("DigitalIODriver:  problemas con sockets");
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
		}
		catch (Exception e)
		{
			System.out.println("Nueva exception");
			e.printStackTrace();
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
		}

		// Creamos el array de OutputGenerator
		generators = new OutputGenerator[NUM_LINE];
		

		
		
	}

	// Retorna true si la l�nea digital de entrada de �ndice line est�
	// establecida a 1
	// l�gico, y false en el caso contrario.
	public boolean readInputLine(int line) throws IndexOutOfRange {
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();
		long inputLines;
		inputLines = theCard.readInputBuffer();
		return ((inputLines & (1L << line)) != 0);
	}

	// Establece la l�nea de salida de �ndice line al estado l�gico state,
	// dejando las
	// 31 l�neas restantes sin cambiar. Dado que el estado de las l�neas de
	// salida no
	// pueden leerse de la tarjeta, el estado de la salida debe mantenerse como
	// una
	// variable de estado en el driver. Si en la l�nea se estuviese generando un
	// pulso
	// o parpadeo, debe ser cancelados.
	public void writeOutputLine(int line, boolean state, Ice.Current curr) throws IndexOutOfRange {
		if (generators[line] != null) {
			generators[line].finish();
			generators[line] = null;
		}
		internalwriteOutputLine(line, state);

	}

	// M�todo privado
	private void internalwriteOutputLine(int line, boolean state)
			throws IndexOutOfRange {
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();

		if (state)
			outputCurrentState = outputCurrentState | (1L << line);
		else
			outputCurrentState = outputCurrentState & ~(1L << line);

		theCard.writeOutputBuffer(outputCurrentState);
	}

	// Retorna True si la l�nea de salida de indice line est� establecida,
	// y false en caso contrario. Lanza la excepci�n IndexOutOfRange si line
	// est� fuera del intervalo 0..31.
	public synchronized boolean readOutputLine(int line) {
		return ((outputCurrentState & (1L << line)) != 0);
	}

	// Suspende el thread que invoca el m�todo, hasta que en la l�nea de entrada
	// de
	// �ndice line presenta una transici�n del estado false al estado true.
	synchronized public void awaitChangeFalseToTrue(int line,
						Ice.Current curr) throws IndexOutOfRange, MeHanInterrumpido
						
	{
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();		
		while (readInputLine(line)) {
			try {
				this.wait();
			} catch (InterruptedException e) {
				throw new MeHanInterrumpido();
			}
		}
		while (!readInputLine(line)) {
			try {
				this.wait();
			} catch (InterruptedException e) {
				throw new MeHanInterrumpido();
			}
		}		

	}

	// Suspende el thread que invoca el m�todo hasta que la l�nea de entrada de
	// �ndice line est� en el estado true.
	synchronized public void awaitTrue(int line) throws InterruptedException, IndexOutOfRange {
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();		
		while (!readInputLine(line)) {
			this.wait();
		}		
	}

	// Comienza a generar una se�al cuadrada en la l�nea de salida de �ndice
	// line con
	// period/2 milisegundos en cada estado. Si se invoca con period=0
	// representa la
	// cancelaci�n de la generaci�n de la se�al cuadrada en la correspondiente
	// l�nea.
	public void blinking(int line, int period, Ice.Current curr) throws IndexOutOfRange {
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();
		// No implementada
		if (generators[line] != null) {
			generators[line].finish();
		}
		generators[line] = new BlinkingGenerator(line, period);

	}

	// Genera en la l�nea de salida de �ndice line un pulso de anchura width
	// milisegundos,
	// esto es, con independencia del estado en que se encuentre esa l�nea,
	// pasar� o per-
	// manecer� en ese estado width milisegundos, y luego pasar� a estado false.
	public void generatePulse(int line, int width, Ice.Current curr) throws IndexOutOfRange {
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();

		if (generators[line] != null) {
			generators[line].finish();
		}
		this.writeOutputLine(line, true, null);
		generators[line] = new PulseGenerator(line, width);
        
	}

	/**
	 * +++++++ M�todos requeridos por la interfaz DigitalIOListener ++++++
	 */
	// M�todo que ser� invocado cada vez que cambie una entrada digital
	// de la tarjeta DigitalIOCard.
	synchronized public void newChange(Ice.Current curr) {		
		notifyAll();
	}

	// Clases internas
	/**
	 * Clase generadora de salida
	 */
	public abstract class OutputGenerator extends Thread {
		int line;
		boolean terminado;

		OutputGenerator(int line) {
			this.line = line;
		}

		public void finish() {
			this.interrupt();
		}
	}

	/**
	 * Clase generadora de pulsos cuadrados
	 */

	private class PulseGenerator extends OutputGenerator {
		long width;

		PulseGenerator(int line, long width) {
			super(line);
			this.width = width;
			this.start();
		}

		public void run() {
//			try {
//				internalwriteOutputLine(line, true);
//			} catch (IndexOutOfRange e1) {}
			try {
				Thread.sleep(width);
				try {
					internalwriteOutputLine(line, false);
				} catch (IndexOutOfRange e) {
					return;
				}
			} catch (InterruptedException e) {
				return;
			}
		}
	}

	/**
	 * Clase generadora de blinking
	 */

	public class BlinkingGenerator extends OutputGenerator {

		long period;

		BlinkingGenerator(int line, long period) {
			super(line);
			this.period = period;
			this.line = line;
			this.start();
		}

		public void run() {
			while (!this.isInterrupted()) {
				boolean value = readOutputLine(line);
				try {
					internalwriteOutputLine(line, !value);
				} catch (IndexOutOfRange e1) {
				}
				try {
					Thread.sleep(period / 2);
				} catch (InterruptedException e) {
					return;
				}
			}
		}
	}
}