import DigitalIOCardPkt.DigitalIOCardRmtPrx;
import DigitalIOCardPkt.DigitalIOCardRmtPrxHelper;
import DigitalIOCardPkt.DigitalIOListenerRmtPrx;
import DigitalIOCardPkt.DigitalIOListenerRmtPrxHelper;
import DigitalIOCardPkt._DigitalIOListenerRmtOperations;
import DigitalIOCardPkt._DigitalIOListenerRmtTie;



/**********************************************************************
 * Programaci�n Concurrente y Distribuida
 * Ingenier�a Inform�tica - Universidad de Cantabria
 * 
 * Proyecto: Clase DigitalIO Driver
 * 
 * Fichero: Clase ActiveIODigitalDriver
 * 
 * @Autor: J.M. Drake y L.Barros
 * @Fecha: 22-10-08
 * 
 *********************************************************************/

/**
 * Driver que gestiona el acceso concurrente a una tarjeta de entrada/salida
 * digital. Implementa la interfaz DigitalIOListener
 */
public class DigitalIODriver implements _DigitalIOListenerRmtOperations {

	// Estado actual de las lineas digitales de salida. Se requiere como
	// variable de estado del driver porque la tajeta no permite leer el
	// valor que est� establecida en la tarjeta.
	public long outputCurrentState;

	// Referencia la tarjeta de IO Digital que gestiona.
	private DigitalIOCardRmtPrx theCard;

	private final int NUM_LINE = 32;

	OutputGenerator[] generators;

	/**
	 * Constructor de la clase
	 */
	DigitalIODriver(Ice.Communicator ic)
	{
		this.outputCurrentState = 0;
		
		String rmtname = DatosComunes.cardRMIname;
		int puerto_tcp = DatosComunes.tcp_port_card;
		String icestr = rmtname + ":default -p " + puerto_tcp;
		
		Ice.ObjectPrx proxy_general = null;
		
		boolean exito = false;
		
		while (!exito)
		{
			try 
			{
				System.out.println("DigitalIODriver:  Buscando " + rmtname  + 
						" en puerto TCP " + puerto_tcp);

				proxy_general = ic.stringToProxy(icestr);
				System.out.println("Proxy creado, proxy_general: " + proxy_general 
									+ " ahora haciendo el cast");
				
				theCard = DigitalIOCardRmtPrxHelper.checkedCast(proxy_general);

				System.out.println("DigitalIODriver:  encuentra DigitalIOCard");
				exito = true;
			}
			catch (Ice.ConnectFailedException e)
			{
				// El lado remoto no está listo. Repetimos hasta que lo esté
				System.out.println("Driver espera a que la tarjeta  " + rmtname + " este escuchando en el puerto " + puerto_tcp);
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}

			}
			catch (Exception e)
			{
				System.out.println("excepcion nueva" + e);
				e.printStackTrace();
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
			}

			
			
		}
		
		/* Exportamos un stub como callback */
		/************************************/
		
		puerto_tcp = DatosComunes.tcp_port_io_listener;
		icestr = "default -p " + puerto_tcp;

		// Al pasar el proxy como callback, el nombre con el que se registra
		// en el Ice adapter no hace falta compartirlo con la clase cliente
		String nombre_local_no_hace_falta_exportar = "NOMBREPUROLOCAL";
		
		Ice.ObjectAdapter adapter = null;
		DigitalIOListenerRmtPrx stub = null;
		try
		{
			adapter = ic.createObjectAdapterWithEndpoints("digitalIOlistener", icestr);
			_DigitalIOListenerRmtTie servant = new _DigitalIOListenerRmtTie(this);
		
			// Tomamos el valor retornado por adapter.add() para generar de ahi el stub
			// que pasaremos en la funcion de paso.
			proxy_general = adapter.add(servant, ic.stringToIdentity(nombre_local_no_hace_falta_exportar) );
			stub = DigitalIOListenerRmtPrxHelper.checkedCast(proxy_general);
			adapter.activate();
		}
		catch (Ice.SocketException e)
		{
			System.out.println("SocketException:  puerto_tcp " + puerto_tcp + "ocupado");
			System.exit(-1);
		}
		catch (Ice.AlreadyRegisteredException e)
		{
			// Esto ocurre por 2 razones:
			// -  ic.createObjectAdapterWithEndpoints:  Un adapter ya esta creado con el mismo
			//                                       nombre
			//
			// -  adapter.add:  Un stub esta registrado con el mismo nombre en el adapter
			System.out.println("Nombre de interfaz " + nombre_local_no_hace_falta_exportar + " ocupado");
			System.exit(-1);
		}
		catch (Exception e)
		{
			System.out.println("Salto otra excepcion");
			e.printStackTrace();
		}

		
		
		theCard.addInterruptListener(stub);
		
		// Creamos el array de OutputGenerator
		generators = new OutputGenerator[NUM_LINE];
	}

	
	
	
	// Retorna true si la l�nea digital de entrada de �ndice line est�
	// establecida a 1
	// l�gico, y false en el caso contrario.
	public boolean readInputLine(int line) throws IndexOutOfRange {
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();
		long inputLines = -1;
			inputLines = theCard.readInputBuffer();

		return ((inputLines & (1L << line)) != 0);
	}

	// Establece la l�nea de salida de �ndice line al estado l�gico state,
	// dejando las
	// 31 l�neas restantes sin cambiar. Dado que el estado de las l�neas de
	// salida no
	// pueden leerse de la tarjeta, el estado de la salida debe mantenerse como
	// una
	// variable de estado en el driver. Si en la l�nea se estuviese generando un
	// pulso
	// o parpadeo, debe ser cancelados.
	public void writeOutputLine(int line, boolean state) throws IndexOutOfRange {
		if (generators[line] != null) {
			generators[line].finish();
			generators[line] = null;
		}
		internalwriteOutputLine(line, state);

	}

	// M�todo privado
	private void internalwriteOutputLine(int line, boolean state)
			throws IndexOutOfRange {
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();

		if (state)
			outputCurrentState = outputCurrentState | (1L << line);
		else
			outputCurrentState = outputCurrentState & ~(1L << line);

		theCard.writeOutputBuffer(outputCurrentState);
	}

	// Retorna True si la l�nea de salida de indice line est� establecida,
	// y false en caso contrario. Lanza la excepci�n IndexOutOfRange si line
	// est� fuera del intervalo 0..31.
	public synchronized boolean readOutputLine(int line) {
		return ((outputCurrentState & (1L << line)) != 0);
	}

	// Suspende el thread que invoca el m�todo, hasta que en la l�nea de entrada
	// de
	// �ndice line presenta una transici�n del estado false al estado true.
	synchronized public void awaitChangeFalseToTrue(int line) throws InterruptedException, IndexOutOfRange {
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();		
		while (readInputLine(line)) {
			this.wait();
		}
		while (!readInputLine(line)) {
			this.wait();
		}		

	}

	// Suspende el thread que invoca el m�todo hasta que la l�nea de entrada de
	// �ndice line est� en el estado true.
	synchronized public void awaitTrue(int line) throws InterruptedException, IndexOutOfRange {
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();		
		while (!readInputLine(line)) {
			this.wait();
		}		
	}

	// Comienza a generar una se�al cuadrada en la l�nea de salida de �ndice
	// line con
	// period/2 milisegundos en cada estado. Si se invoca con period=0
	// representa la
	// cancelaci�n de la generaci�n de la se�al cuadrada en la correspondiente
	// l�nea.
	public void blinking(int line, int period) throws IndexOutOfRange {
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();
		// No implementada
		if (generators[line] != null) {
			generators[line].finish();
		}
		generators[line] = new BlinkingGenerator(line, period);

	}

	// Genera en la l�nea de salida de �ndice line un pulso de anchura width
	// milisegundos,
	// esto es, con independencia del estado en que se encuentre esa l�nea,
	// pasar� o per-
	// manecer� en ese estado width milisegundos, y luego pasar� a estado false.
	public void generatePulse(int line, int width) throws IndexOutOfRange {
		if (line < 0 || line > 31)
			throw new IndexOutOfRange();

		if (generators[line] != null) {
			generators[line].finish();
		}
		this.writeOutputLine(line, true);
		generators[line] = new PulseGenerator(line, width);
        
	}

	/**
	 * +++++++ M�todos requeridos por la interfaz DigitalIOListener ++++++
	 */
	// M�todo que ser� invocado cada vez que cambie una entrada digital
	// de la tarjeta DigitalIOCard.
	synchronized public void newChange(Ice.Current curr) {		
		notifyAll();
	}

	// Clases internas
	/**
	 * Clase generadora de salida
	 */
	public abstract class OutputGenerator extends Thread {
		int line;
		boolean terminado;

		OutputGenerator(int line) {
			this.line = line;
		}

		public void finish() {
			this.interrupt();
		}
	}

	/**
	 * Clase generadora de pulsos cuadrados
	 */

	private class PulseGenerator extends OutputGenerator {
		long width;

		PulseGenerator(int line, long width) {
			super(line);
			this.width = width;
			this.start();
		}

		public void run() {
//			try {
//				internalwriteOutputLine(line, true);
//			} catch (IndexOutOfRange e1) {}
			try {
				Thread.sleep(width);
				try {
					internalwriteOutputLine(line, false);
				} catch (IndexOutOfRange e) {
					return;
				}
			} catch (InterruptedException e) {
				return;
			}
		}
	}

	/**
	 * Clase generadora de blinking
	 */

	public class BlinkingGenerator extends OutputGenerator {

		long period;

		BlinkingGenerator(int line, long period) {
			super(line);
			this.period = period;
			this.line = line;
			this.start();
		}

		public void run() {
			while (!this.isInterrupted()) {
				boolean value = readOutputLine(line);
				try {
					internalwriteOutputLine(line, !value);
				} catch (IndexOutOfRange e1) {
				}
				try {
					Thread.sleep(period / 2);
				} catch (InterruptedException e) {
					return;
				}
			}
		}
	}
}